
/*
 *@@sourcefile xfdisk.c:
 *      This file contains SOM code for the following XWorkplace classes:
 *
 *      --  XFldDisk (WPDisk replacement)
 *
 *      XFldDisk is needed mainly to modify the popup menus
 *      of Disk objects also. Since these are not instances
 *      of WPFolder, we need an extra subclass.
 *
 *      Installation of this class is now optional (V0.9.0).
 *      However, if it is installed, XFolder must also be
 *      installed.
 *
 *      Starting with V0.9.0, the files in classes\ contain only
 *      i.e. the methods themselves.
 *      The implementation for this class is mostly in filesys\disk.c.
 *
 *@@somclass XFldDisk xfdisk_
 *@@somclass M_XFldDisk xfdiskM_
 */

/*
 *      Copyright (C) 1997-2000 Ulrich Mller.
 *      This file is part of the XWorkplace source package.
 *      XWorkplace is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published
 *      by the Free Software Foundation, in version 2 as it comes in the
 *      "COPYING" file of the XWorkplace main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

/*
 *  This file was generated by the SOM Compiler and Emitter Framework.
 *  Generated using:
 *      SOM Emitter emitctm: 2.42
 */

#ifndef SOM_Module_xfdisk_Source
#define SOM_Module_xfdisk_Source
#endif
#define XFldDisk_Class_Source
#define M_XFldDisk_Class_Source

#pragma strings(readonly)

/*
 *  Suggested #include order:
 *  1)  os2.h
 *  2)  C library headers
 *  3)  setup.h (code generation and debugging options)
 *  4)  headers in helpers\
 *  5)  at least one SOM implementation header (*.ih)
 *  6)  dlgids.h, headers in shared\ (as needed)
 *  7)  headers in implementation dirs (e.g. filesys\, as needed)
 *  8)  #pragma hdrstop and then more SOM headers which crash with precompiled headers
 */

#define INCL_DOSSEMAPHORES
#define INCL_DOSERRORS

#define INCL_WINMENUS
#include  <os2.h>

// C library headers
#include <stdio.h>

// generic headers
#include "setup.h"                      // code generation and debugging options

// headers in /helpers
#include "helpers\dosh.h"               // Control Program helper routines
#include "helpers\winh.h"               // PM helper routines

// SOM headers which don't crash with prec. header files
#include "xfdisk.ih"
#include "xfldr.ih"

// XWorkplace implementation headers
#include "dlgids.h"                     // all the IDs that are shared with NLS
#include "shared\common.h"              // the majestic XWorkplace include file
#include "shared\helppanels.h"          // all XWorkplace help panel IDs
#include "shared\kernel.h"              // XWorkplace Kernel
#include "shared\notebook.h"            // generic XWorkplace notebook handling
#include "shared\wpsh.h"                // some pseudo-SOM functions (WPS helper routines)

#include "filesys\disk.h"               // XFldDisk implementation
#include "filesys\folder.h"             // XFolder implementation
#include "filesys\fdrmenus.h"           // shared folder menu logic
#include "filesys\statbars.h"           // status bar translation logic

// other SOM headers
#include "helpers\undoc.h"              // some undocumented stuff
#pragma hdrstop

/* ******************************************************************
 *
 *   Global variables
 *
 ********************************************************************/

// key for wpRestoreData etc.
static const char*  G_pcszXFldDisk = "XFldDisk";

/* ******************************************************************
 *                                                                  *
 *   here come the XFldDisk instance methods                        *
 *                                                                  *
 ********************************************************************/

/*
 *@@ wpFilterPopupMenu:
 *      this WPObject instance method allows the object to
 *      filter out unwanted menu items from the context menu.
 *      This gets called before wpModifyPopupMenu.
 *
 *      We remove default entries for disks according to the
 *      global menu settings.
 *
 *@@changed V0.9.5 (2000-09-18) [umoeller]: now removing "Create another" always
 */

SOM_Scope ULONG  SOMLINK xfdisk_wpFilterPopupMenu(XFldDisk *somSelf,
                                                ULONG ulFlags,
                                                HWND hwndCnr,
                                                BOOL fMultiSelect)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    // XFldDiskData *somThis = XFldDiskGetData(somSelf);
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpFilterPopupMenu");

    return (XFldDisk_parent_WPDisk_wpFilterPopupMenu(somSelf,
                                                     ulFlags,
                                                     hwndCnr,
                                                     fMultiSelect)
        & ~( pGlobalSettings->DefaultMenuItems | CTXT_NEW )
        );
}

/*
 *@@ wpModifyPopupMenu:
 *      this WPObject instance methods gets called by the WPS
 *      when a context menu needs to be built for the object
 *      and allows the object to manipulate its context menu.
 *      This gets called _after_ wpFilterPopupMenu.
 *
 *      We add the various XFolder menu entries here
 *      by calling mnuModifyFolderPopupMenu in fdrmenus.c,
 *      which is also used by the XFolder class.
 *
 *@@todo hwndCnr is NULLHANDLE always
 */

SOM_Scope BOOL  SOMLINK xfdisk_wpModifyPopupMenu(XFldDisk *somSelf,
                                                 HWND hwndMenu,
                                                 HWND hwndCnr,
                                                 ULONG iPosition)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    BOOL            rc;

    // get the "root folder" of the WPDisk object; that is the
    // folder which is opened when the disk object is opened
    // (i.e. the folder of the root directory). This returns
    // NULL if the drive is not ready.
    XFolder         *pFolder = _wpQueryRootFolder(somSelf);

    // XFldDiskData *somThis = XFldDiskGetData(somSelf);
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpModifyPopupMenu");

    rc = XFldDisk_parent_WPDisk_wpModifyPopupMenu(somSelf,hwndMenu,hwndCnr,iPosition);

    if (pGlobalSettings->RemoveFormatDiskItem)
        winhRemoveMenuItem(hwndMenu, ID_WPMI_FORMATDISK);

    if (pGlobalSettings->RemoveCheckDiskItem)
        winhRemoveMenuItem(hwndMenu, ID_WPMI_CHECKDISK);

    if (pFolder)
        // drive ready:
        rc = mnuModifyFolderPopupMenu(pFolder,
                                      hwndMenu,
                                      hwndCnr,    // @@todo this seems to be NULLHANDLE
                                      iPosition);

    return (rc);
}

/*
 *@@ wpMenuItemSelected:
 *      this WPObject method processes menu selections.
 *      This must be overridden to support new menu
 *      items which have been added in wpModifyPopupMenu.
 *
 *      See XFldObject::wpMenuItemSelected for additional
 *      remarks.
 *
 *      We pass the input to mnuMenuItemSelected in fdrmenus.c
 *      because disk menu items are mostly shared with XFolder.
 */

SOM_Scope BOOL  SOMLINK xfdisk_wpMenuItemSelected(XFldDisk *somSelf,
                                                  HWND hwndFrame,
                                                  ULONG ulMenuId)
{
    XFolder         *pFolder = _wpQueryRootFolder(somSelf);
    // XFldDiskData *somThis = XFldDiskGetData(somSelf);
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpMenuItemSelected");

    if (mnuMenuItemSelected(pFolder, hwndFrame, ulMenuId))
        return TRUE;
    else
        return (XFldDisk_parent_WPDisk_wpMenuItemSelected(somSelf, hwndFrame,
                                                   ulMenuId));
}

/*
 *@@ wpMenuItemHelpSelected:
 *      display help for a context menu item;
 *      we pass the input to mnuMenuItemHelpSelected in fdrmenus.c.
 */

SOM_Scope BOOL  SOMLINK xfdisk_wpMenuItemHelpSelected(XFldDisk *somSelf,
                                                      ULONG MenuId)
{
    XFolder         *pFolder = _wpQueryRootFolder(somSelf);
    // XFldDiskData *somThis = XFldDiskGetData(somSelf);
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpMenuItemHelpSelected");

    if (mnuMenuItemHelpSelected(pFolder, MenuId))
        return TRUE;
    else
        return (XFldDisk_parent_WPDisk_wpMenuItemHelpSelected(somSelf,
                                                           MenuId));
}

/*
 *@@ wpViewObject:
 *      this WPObject method either opens a new view of the
 *      object (by calling wpOpen) or resurfaces an already
 *      open view, if one exists already and "concurrent views"
 *      are enabled. This gets called every single time when
 *      an object is to be opened... e.g. on double-clicks.
 *
 *      For WPDisks, the WPS seems to be doing no drive checking
 *      in here, which leads to the annoying "Drive not ready"
 *      popups. So we try to implement this here.
 *
 *@@changed V0.9.0 [umoeller]: added global setting for disabling this feature
 */

SOM_Scope HWND  SOMLINK xfdisk_wpViewObject(XFldDisk *somSelf,
                                            HWND hwndCnr,
                                            ULONG ulView,
                                            ULONG param)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    HWND            hwndFrame = NULLHANDLE; // default: error occured

    /* XFldDiskData *somThis = XFldDiskGetData(somSelf); */
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpViewObject");

    // "Drive not ready" replacement enabled?
    if (    (pGlobalSettings->fReplDriveNotReady)
         && (ulView != OPEN_SETTINGS)
       )
    {
        // yes: use the safe way of opening the
        // drive (this prompts the user upon errors)
        XFolder*        pRootFolder = NULL;
        pRootFolder = dskCheckDriveReady(somSelf);
        if (pRootFolder)
            // success: call default
            hwndFrame = XFldDisk_parent_WPDisk_wpViewObject(somSelf, hwndCnr,
                                                            ulView, param);
    }
    else
        hwndFrame = XFldDisk_parent_WPDisk_wpViewObject(somSelf, hwndCnr,
                                                        ulView, param);

    return (hwndFrame);
}

/*
 *@@ wpOpen:
 *      this WPObject instance method gets called when
 *      a new view needs to be opened. Normally, this
 *      gets called after wpViewObject has scanned the
 *      object's USEITEMs and has determined that a new
 *      view is needed.
 *
 *      This _normally_ runs on thread 1 of the WPS, but
 *      this is not always the case. If this gets called
 *      in response to a menu selection from the "Open"
 *      submenu or a double-click in the folder, this runs
 *      on the thread of the folder (which _normally_ is
 *      thread 1). However, if this results from WinOpenObject
 *      or an OPEN setup string, this will not be on thread 1.
 *
 *      We call the parent first and then subclass the
 *      resulting frame window, similar to what we're doing
 *      with folder views (see XFolder::wpOpen).
 *
 *@@changed V0.9.2 (2000-03-06) [umoeller]: drives were checked even if replacement dlg was disabled; fixed
 *@@changed V0.9.3 (2000-04-08) [umoeller]: adjusted for new folder subclassing
 */

SOM_Scope HWND  SOMLINK xfdisk_wpOpen(XFldDisk *somSelf,
                                      HWND hwndCnr,
                                      ULONG ulView,
                                      ULONG param)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    HWND            hwndNewFrame = NULLHANDLE; // default: error occured

    switch (ulView)
    {
        case OPEN_CONTENTS:
        case OPEN_TREE:
        case OPEN_DETAILS:
        {
            APIRET          arc = NO_ERROR;
            XFolder*        pRootFolder = 0;
            // XFldDiskData *somThis = XFldDiskGetData(somSelf);
            XFldDiskMethodDebug("XFldDisk","xfdisk_wpOpen");

            // "Drive not ready" replacement enabled?
            if (pGlobalSettings->fReplDriveNotReady)
            {
                // query root folder (WPRootFolder class, which is a descendant
                // of WPFolder/XFolder); each WPDisk is paired with one of those,
                // and the actual display is done by this class, so we will pass
                // pRootFolder instead of somSelf to most following method calls.
                // We use wpshQueryRootFolder instead of wpQueryRootFolder to
                // avoid "Drive not ready" popups.
                pRootFolder = wpshQueryRootFolder(somSelf, FALSE, &arc);
                if (pRootFolder)
                    // drive ready: call parent to get frame handle
                    hwndNewFrame = XFldDisk_parent_WPDisk_wpOpen(somSelf,
                                                                 hwndCnr,
                                                                 ulView,
                                                                 param);
                // else: hwndNewFrame is still NULLHANDLE
            }
            else
            {
                // "drive not ready" replacement disabled:
                hwndNewFrame = XFldDisk_parent_WPDisk_wpOpen(somSelf,
                                                             hwndCnr,
                                                             ulView,
                                                             param);
                if (hwndNewFrame)
                    // no error:
                    pRootFolder = _wpQueryRootFolder(somSelf);
            }

            if (pRootFolder)
            {

                if (hwndNewFrame)
                {
                    // open successful:
                    if (   (ulView == OPEN_CONTENTS)
                        || (ulView == OPEN_TREE)
                        || (ulView == OPEN_DETAILS)
                       )
                    {
                        PSUBCLASSEDFOLDERVIEW psfv = NULL;

                        hwndCnr = wpshQueryCnrFromFrame(hwndNewFrame);

                        // subclass frame window; this is the same
                        // proc which is used for normal folder frames,
                        // we just use pRootFolder instead.
                        // However, we pass somSelf as the "real" object
                        // which will then be stored in *psli.
                        psfv = fdrSubclassFolderView(hwndNewFrame,
                                                     hwndCnr,
                                                     pRootFolder, // folder
                                                     somSelf);    // real object; for disks, this
                                                                  // is the WPDisk object...
                        // add status bar, if allowed: as opposed to
                        // XFolder's, for XFldDisk's we only check the
                        // global setting, because there's no instance
                        // setting for this with XFldDisk's
                        if (    (pGlobalSettings->fEnableStatusBars)
                                                        // feature enabled?
                             && (pGlobalSettings->fDefaultStatusBarVisibility)
                                                        // bars visible per default?
                           )
                            // assert that subclassed list item is valid
                            if (psfv)
                                // add status bar only if allowed for the current view type
                                if (    (   (ulView == OPEN_CONTENTS)
                                         && (pGlobalSettings->SBForViews & SBV_ICON)
                                        )
                                     || (   (ulView == OPEN_TREE)
                                         && (pGlobalSettings->SBForViews & SBV_TREE)
                                        )
                                     || (   (ulView == OPEN_DETAILS)
                                         && (pGlobalSettings->SBForViews & SBV_DETAILS)
                                        )
                                    )
                                    // this reformats the window with the status bar
                                    fdrCreateStatusBar(pRootFolder, psfv, TRUE);

                        // extended sort functions
                        if (pGlobalSettings->ExtFolderSort)
                            if (hwndCnr)
                                fdrSetFldrCnrSort(pRootFolder,
                                                  hwndCnr,
                                                  FALSE);
                    }
                }
            } // if (pRootFolder)
        break; }

        default:
            // e.g. settings view
            hwndNewFrame = XFldDisk_parent_WPDisk_wpOpen(somSelf,
                                                         hwndCnr,
                                                         ulView,
                                                         param);
    } // switch (ulView)

    return (hwndNewFrame);
}

/*
 *@@ wpAddDiskDetailsPage:
 *      this adds the "Details" page to a disk object's
 *      settings notebook.
 *
 *      We override this method to insert our enhanced
 *      page instead.
 *
 *@@added V0.9.0 [umoeller]
 */

SOM_Scope ULONG  SOMLINK xfdisk_wpAddDiskDetailsPage(XFldDisk *somSelf,
                                                     HWND hwndNotebook)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    /* XFldDiskData *somThis = XFldDiskGetData(somSelf); */
    XFldDiskMethodDebug("XFldDisk","xfdisk_wpAddDiskDetailsPage");

    if (pGlobalSettings->fReplaceFilePage)
    {
        PCREATENOTEBOOKPAGE pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));

        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndNotebook;
        pcnbp->hmod = cmnQueryNLSModuleHandle(FALSE);
        pcnbp->ulDlgID = ID_XSD_DISK_DETAILS;
        pcnbp->ulPageID = SP_DISK_DETAILS;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->pszName = cmnGetString(ID_XSSI_DETAILSPAGE);  // pszDetailsPage
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_DISKDETAILS;

        pcnbp->pfncbInitPage    = (PFNCBACTION)dskDetailsInitPage;
        pcnbp->pfncbItemChanged = dskDetailsItemChanged;

        return (ntbInsertPage(pcnbp));
    }
    else
        return (XFldDisk_parent_WPDisk_wpAddDiskDetailsPage(somSelf,
                                                            hwndNotebook));
}

/* ******************************************************************
 *                                                                  *
 *   here come the XFldDisk class methods                           *
 *                                                                  *
 ********************************************************************/

/*
 *@@ wpclsInitData:
 *      initialize XFldDisk class data.
 *
 *@@changed V0.9.0 [umoeller]: added class object to KERNELGLOBALS
 */

SOM_Scope void  SOMLINK xfdiskM_wpclsInitData(M_XFldDisk *somSelf)
{
    // M_XFldDiskData *somThis = M_XFldDiskGetData(somSelf);
    M_XFldDiskMethodDebug("M_XFldDisk","xfdiskM_wpclsInitData");

    M_XFldDisk_parent_M_WPDisk_wpclsInitData(somSelf);

    {
        // store the class object in KERNELGLOBALS
        PKERNELGLOBALS   pKernelGlobals = krnLockGlobals(__FILE__, __LINE__, __FUNCTION__);
        if (pKernelGlobals)
        {
            pKernelGlobals->fXFldDisk = TRUE;
            krnUnlockGlobals();
        }
    }
}

/*
 *@@ wpclsQueryTitle:
 *      this WPObject class method tells the WPS the clear
 *      name of a class, which is shown in the third column
 *      of a Details view and also used as the default title
 *      for new objects of a class.
 *
 *      We override the standard folder class name only if
 *      the user has enabled "fix class titles" in XWPSetup.
 *
 *@@added V0.9.12 (2001-05-22) [umoeller]
 */

SOM_Scope PSZ  SOMLINK xfdiskM_wpclsQueryTitle(M_XFldDisk *somSelf)
{
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    /* M_XFldDiskData *somThis = M_XFldDiskGetData(somSelf); */
    M_XFldDiskMethodDebug("M_XFldDisk","xfdiskM_wpclsQueryTitle");

    if (pGlobalSettings->fFixClassTitles)
        return (cmnGetString(ID_XSSI_CLASSTITLE_DISK));

    return (M_XFldDisk_parent_M_WPDisk_wpclsQueryTitle(somSelf));
}


