
/*
 *@@sourcefile xfwps.c:
 *      This file contains SOM code for the following XWorkplace classes:
 *
 *      --  XFldWPS SOM stuff (WPSystem subclass)
 *
 *      This class implements the "Workplace Shell" settings
 *      object.
 *
 *      Installation of this class is optional, but you won't
 *      be able to change XWorkplace's settings without it.
 *
 *      This class used to be in xfsys.c (together with XFldSystem,
 *      the "OS/2 Kernel" object), but has been moved here for
 *      clarity with V0.9.0.
 *
 *      Starting with V0.9.0, the files in classes\ contain only
 *      i.e. the methods themselves.
 *      The implementation for this class is in several files in
 *      filesys\; check the function prefixes to find out where.
 *
 *@@somclass XFldWPS xfwps_
 *@@somclass M_XFldWPS xfwpsM_
 */

/*
 *      Copyright (C) 1997-2000 Ulrich Mller.
 *      This file is part of the XWorkplace source package.
 *      XWorkplace is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published
 *      by the Free Software Foundation, in version 2 as it comes in the
 *      "COPYING" file of the XWorkplace main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

/*
 *  This file was generated by the SOM Compiler and Emitter Framework.
 *  Generated using:
 *      SOM Emitter emitctm: 2.41
 */

#ifndef SOM_Module_xfwps_Source
#define SOM_Module_xfwps_Source
#endif
#define xfwps_Class_Source
#define M_xfwps_Class_Source

#pragma strings(readonly)

/*
 *  Suggested #include order:
 *  1)  os2.h
 *  2)  C library headers
 *  3)  setup.h (code generation and debugging options)
 *  4)  headers in helpers\
 *  5)  at least one SOM implementation header (*.ih)
 *  6)  dlgids.h, headers in shared\ (as needed)
 *  7)  headers in implementation dirs (e.g. filesys\, as needed)
 *  8)  #pragma hdrstop and then more SOM headers which crash with precompiled headers
 */

#define INCL_DOSSEMAPHORES
#define INCL_WINMENUS
#include <os2.h>

// C library headers
#include <stdio.h>              // needed for except.h

// generic headers
#include "setup.h"                      // code generation and debugging options

// headers in /helpers
#include "helpers\dosh.h"               // Control Program helper routines

// SOM headers which don't crash with prec. header files
#include "xfwps.ih"
#include "xfldr.ih"

// XWorkplace implementation headers
#include "dlgids.h"                     // all the IDs that are shared with NLS
#include "shared\common.h"              // the majestic XWorkplace include file
#include "shared\helppanels.h"          // all XWorkplace help panel IDs
#include "shared\kernel.h"              // XWorkplace Kernel
#include "shared\notebook.h"            // generic XWorkplace notebook handling

#include "filesys\filetype.h"           // extended file types implementation
#include "filesys\folder.h"             // XFolder implementation
#include "filesys\fdrmenus.h"           // shared folder menu logic
#include "filesys\statbars.h"           // status bar translation logic

// other SOM headers
#pragma hdrstop                         // VAC++ keeps crashing otherwise

/* ******************************************************************
 *                                                                  *
 *   here come the XFldWPS instance methods                         *
 *                                                                  *
 ********************************************************************/

/*
 *@@ xwpAddXFldWPSPages:
 *      this actually adds the XWorkplace pages into the
 *      "Workplace Shell" notebook.
 *
 *@@changed V0.9.0 [umoeller]: added "File types" page, removed "XFolder" pages
 *@@changed V0.9.6 (2000-10-16) [umoeller]: made "File types" resizeable
 */

SOM_Scope ULONG  SOMLINK xfwps_xwpAddXFldWPSPages(XFldWPS *somSelf,
                                                 HWND hwndDlg)
{
    PCREATENOTEBOOKPAGE pcnbp;
    HMODULE         savehmod = cmnQueryNLSModuleHandle(FALSE);
    PCGLOBALSETTINGS pGlobalSettings = cmnQueryGlobalSettings();
    // XFldWPSData *somThis = XFldWPSGetData(somSelf);
    XFldWPSMethodDebug("XFldWPS","xfwps_xwpAddXFldWPSPages");

    // insert "Sort" page
    if (pGlobalSettings->ExtFolderSort)
    {
        // extended sorting enabled:
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->ulDlgID = ID_XSD_SETTINGS_FLDRSORT;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->pszName = cmnGetString(ID_XSSI_SORT);  // pszSort
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SORTPAGE;
                        // changed V0.9.12 (2001-05-20) [umoeller]

        // mark this page as "global", because both
        // the instance settings notebook and the
        // "Workplace Shell" object use the same
        // callbacks
        pcnbp->ulPageID = SP_FLDRSORT_GLOBAL;
        pcnbp->pfncbInitPage    = fdrSortInitPage;
        pcnbp->pfncbItemChanged = fdrSortItemChanged;
        ntbInsertPage(pcnbp);
    }

    // insert "Hotkeys" page
    if (pGlobalSettings->fEnableFolderHotkeys)
    {
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->ulDlgID = ID_XSD_SET4ACCELS;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->pszName = cmnGetString(ID_XSSI_4ACCELERATORS);  // psz4Accelerators
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS1+5;
        pcnbp->ulPageID = SP_4ACCELERATORS;
        pcnbp->pfncbInitPage    = fdrHotkeysInitPage;
        pcnbp->pfncbItemChanged = fdrHotkeysItemChanged;
        ntbInsertPage(pcnbp);
    }

    // insert "Snap to grid" page
    if (pGlobalSettings->fEnableSnap2Grid)
    {
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->pszName = cmnGetString(ID_XSSI_3SNAPTOGRID);  // psz3SnapToGrid
        pcnbp->ulDlgID = ID_XSD_SET3SNAPTOGRID;
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS1+4;
        pcnbp->ulPageID = SP_3SNAPTOGRID;
        pcnbp->pfncbInitPage    = fdrGridInitPage;
        pcnbp->pfncbItemChanged = fdrGridItemChanged;
        ntbInsertPage(pcnbp);
    }

    /*
     * "Status bar" pages
     */

    if (    (!pGlobalSettings->fNoSubclassing)
         && (pGlobalSettings->fEnableStatusBars)
       )
    {
        // insert "Status bar" page 2
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->usPageStyleFlags = BKA_MINOR;
        pcnbp->fEnumerate = TRUE;
        pcnbp->pszName = cmnGetString(ID_XSSI_27STATUSBAR);  // psz27StatusBar
        pcnbp->ulDlgID = ID_XSD_SET28STATUSBARS2;
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_SB2;
        pcnbp->ulPageID = SP_28STATUSBAR2;
        pcnbp->pfncbInitPage    = stbStatusBar2InitPage;
        pcnbp->pfncbItemChanged = stbStatusBar2ItemChanged;
        ntbInsertPage(pcnbp);

        // insert "Status bar" page 1
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->fEnumerate = TRUE;
        pcnbp->pszName = cmnGetString(ID_XSSI_27STATUSBAR);  // psz27StatusBar
        pcnbp->ulDlgID = ID_XSD_SET27STATUSBARS;
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS1+3;
        pcnbp->ulPageID = SP_27STATUSBAR;
        pcnbp->pfncbInitPage    = stbStatusBar1InitPage;
        pcnbp->pfncbItemChanged = stbStatusBar1ItemChanged;
        ntbInsertPage(pcnbp);

        // insert "Status bar" parent major page
        /* pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->fEnumerate = TRUE;
        pcnbp->pszName = cmnGetString(ID_XSSI_27STATUSBAR); psz27StatusBar
        pcnbp->ulDlgID = ID_XSD_SET27STATUSBARS;
        ntbInsertPage(pcnbp); */
    }

    /*
     * "View" page (XFolder)
     */

    pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
    memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
    pcnbp->somSelf = somSelf;
    pcnbp->hwndNotebook = hwndDlg;
    pcnbp->hmod = savehmod;
    pcnbp->usPageStyleFlags = BKA_MAJOR;
    pcnbp->fEnumerate = TRUE;
    pcnbp->pszName = cmnGetString(ID_XSSI_VIEWPAGE);  // pszViewPage
    pcnbp->ulDlgID = ID_XSD_FOLDERVIEWS;
    pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS1;
    pcnbp->ulPageID = SP_1GENERIC;
    pcnbp->pfncbInitPage    = fdrViewInitPage;
    pcnbp->pfncbItemChanged = fdrViewItemChanged;
    ntbInsertPage(pcnbp);

    /*
     * "Menu items" pages
     */

    pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
    memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
    pcnbp->somSelf = somSelf;
    pcnbp->hwndNotebook = hwndDlg;
    pcnbp->hmod = savehmod;
    pcnbp->usPageStyleFlags = BKA_MINOR;
    pcnbp->fEnumerate = TRUE;
    pcnbp->pszName = cmnGetString(ID_XSSI_2REMOVEITEMS);  // psz2RemoveItems
    pcnbp->ulDlgID = ID_XSD_SET2REMOVEMENUS;
    pcnbp->usFirstControlID = ID_XSDI_FIND;
    // pcnbp->ulFirstSubpanel = ID_XSH_SETTINGS_REMOVEMENUS_SUB;        // help panel for "Find"
    pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_REMOVEMENUS;
    pcnbp->ulPageID = SP_2REMOVEITEMS;
    pcnbp->pfncbInitPage    = mnuRemoveMenusInitPage;
    pcnbp->pfncbItemChanged = mnuRemoveMenusItemChanged;
    ntbInsertPage(pcnbp);

    // insert "Config folder menu items" page
    pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
    memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
    pcnbp->somSelf = somSelf;
    pcnbp->hwndNotebook = hwndDlg;
    pcnbp->hmod = savehmod;
    pcnbp->usPageStyleFlags = BKA_MINOR;
    pcnbp->fEnumerate = TRUE;
    pcnbp->pszName = cmnGetString(ID_XSSI_26CONFIGITEMS);  // psz26ConfigFolderMenus
    pcnbp->ulDlgID = ID_XSD_SET26CONFIGMENUS;
    pcnbp->usFirstControlID = ID_XSDI_CASCADE;
    // pcnbp->ulFirstSubpanel = ID_XSH_SETTINGS_CFGM_SUB;       // help panel for "Cascade..."
    pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_CFGM;
    pcnbp->ulPageID = SP_26CONFIGITEMS;
    pcnbp->pfncbInitPage    = mnuConfigFolderMenusInitPage;
    pcnbp->pfncbItemChanged = mnuConfigFolderMenusItemChanged;
    ntbInsertPage(pcnbp);

    // insert "Add Menu Items" page
    pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
    memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
    pcnbp->somSelf = somSelf;
    pcnbp->hwndNotebook = hwndDlg;
    pcnbp->hmod = savehmod;
    pcnbp->usPageStyleFlags = BKA_MAJOR;
    pcnbp->fEnumerate = TRUE;
    pcnbp->pszName = cmnGetString(ID_XSSI_25ADDITEMS);  // psz25AddItems
    pcnbp->ulDlgID = ID_XSD_SET25ADDMENUS;
    pcnbp->usFirstControlID = ID_XSDI_FILEATTRIBS;
    // pcnbp->ulFirstSubpanel = ID_XSH_SETTINGS_ADDMENUS_SUB;   // help panel for "Add file attribs"
    pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_ADDMENUS;
    pcnbp->ulPageID = SP_25ADDITEMS;
    pcnbp->pfncbInitPage    = mnuAddMenusInitPage;
    pcnbp->pfncbItemChanged = mnuAddMenusItemChanged;
    ntbInsertPage(pcnbp);

    /*
     * "File types" page (new with V0.9.0)
     */

    if (pGlobalSettings->fExtAssocs)
    {
        pcnbp = malloc(sizeof(CREATENOTEBOOKPAGE));
        memset(pcnbp, 0, sizeof(CREATENOTEBOOKPAGE));
        pcnbp->somSelf = somSelf;
        pcnbp->hwndNotebook = hwndDlg;
        pcnbp->hmod = savehmod;
        pcnbp->usPageStyleFlags = BKA_MAJOR;
        pcnbp->pszName = cmnGetString(ID_XSSI_FILETYPESPAGE);  // pszFileTypesPage
        pcnbp->ulDlgID = ID_XSD_FILETYPES;
        pcnbp->ulDefaultHelpPanel  = ID_XSH_SETTINGS_FILETYPES;
        pcnbp->ulPageID = SP_FILETYPES;
        pcnbp->pampControlFlags = G_pampFileTypesPage;
        pcnbp->cControlFlags = G_cFileTypesPage;
        pcnbp->pfncbInitPage    = ftypFileTypesInitPage;
        pcnbp->pfncbItemChanged = ftypFileTypesItemChanged;
        ntbInsertPage(pcnbp);
    }

    return (TRUE);
}

/*
 *@@ wpFilterPopupMenu:
 *      this WPObject instance method allows the object to
 *      filter out unwanted menu items from the context menu.
 *      This gets called before wpModifyPopupMenu.
 *
 *      We remove the "Create another" menu item.
 *
 *@@added V0.9.2 (2000-02-26) [umoeller]
 */

SOM_Scope ULONG  SOMLINK xfwps_wpFilterPopupMenu(XFldWPS *somSelf,
                                                 ULONG ulFlags,
                                                 HWND hwndCnr,
                                                 BOOL fMultiSelect)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpFilterPopupMenu");

    return (XFldWPS_parent_WPSystem_wpFilterPopupMenu(somSelf,
                                                      ulFlags,
                                                      hwndCnr,
                                                      fMultiSelect)
            & ~CTXT_NEW
           );
}

/*
 *@@ wpQueryDefaultHelp:
 *      this WPObject instance method specifies the default
 *      help panel for an object (when "Extended help" is
 *      selected from the object's context menu). This should
 *      describe what this object can do in general.
 *      We must return TRUE to report successful completion.
 *
 *      We will display some introduction to "Workplace Shell".
 *
 *@@added V0.9.0 [umoeller]
 */

SOM_Scope BOOL  SOMLINK xfwps_wpQueryDefaultHelp(XFldWPS *somSelf,
                                                 PULONG pHelpPanelId,
                                                 PSZ HelpLibrary)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpQueryDefaultHelp");

    strcpy(HelpLibrary, cmnQueryHelpLibrary());
    *pHelpPanelId = ID_XSH_XFLDWPS;
    return (TRUE);
}

/*
 *@@ wpAddSystemScreenPage:
 *      this WPSystem instance method is overridden in order
 *      to suppress the "Screen" page in the "Workplace Shell"
 *      object, because we want that page in the new "Screen"
 *      object instead.
 */

SOM_Scope ULONG  SOMLINK xfwps_wpAddSystemScreenPage(XFldWPS *somSelf,
                                                     HWND hwndNotebook)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpAddSystemScreenPage");

    return (SETTINGS_PAGE_REMOVED);
}

/*
 *@@ wpAddDMQSDisplayTypePage:
 *      this WPSystem instance method is overridden in order
 *      to suppress the second "Screen" page in the "Workplace
 *      Shell" object. Depending on the installed video driver,
 *      this page may or may not be displayed in the "System"
 *      notebook, but we never want this in "Workplace Shell",
 *      but in "Screen" instead.
 */

SOM_Scope ULONG  SOMLINK xfwps_wpAddDMQSDisplayTypePage(XFldWPS *somSelf,
                                                        HWND hwndNotebook)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpAddDMQSDisplayTypePage");

    return (SETTINGS_PAGE_REMOVED);
}

/*
 *@@ wpAddSystemPrintScreenPage:
 *      this WPSystem instance method is overridden in order
 *      to suppress the "Print screen" page in the "Workplace
 *      Shell" object. We want this page in the new "Screen"
 *      object instead.
 *
 *@@added V0.9.2 (2000-02-23) [umoeller]
 */

SOM_Scope ULONG  SOMLINK xfwps_wpAddSystemPrintScreenPage(XFldWPS *somSelf,
                                                          HWND hwndNotebook)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpAddSystemPrintScreenPage");

    return (SETTINGS_PAGE_REMOVED);
}

/*
 *@@ wpAddSettingsPages:
 *      this WPObject instance method gets called by the WPS
 *      when the Settings view is opened to have all the
 *      settings page inserted into hwndNotebook.
 *
 *      We add the new XWorkplace pages on top of
 *      the other WPS settings pages from the old "System"
 *      notebook.
 */

SOM_Scope BOOL  SOMLINK xfwps_wpAddSettingsPages(XFldWPS *somSelf,
                                                 HWND hwndNotebook)
{
    /* XFldWPSData *somThis = XFldWPSGetData(somSelf); */
    XFldWPSMethodDebug("XFldWPS","xfwps_wpAddSettingsPages");

    XFldWPS_parent_WPSystem_wpAddSettingsPages(somSelf,
                                                       hwndNotebook);

    // add XFolder pages on top
    _xwpAddXFldWPSPages(somSelf, hwndNotebook);

    return (TRUE);
}

/*
 *@@ wpclsInitData:
 *      initialize XFldWPS class data.
 *
 *@@changed V0.9.0 [umoeller]: added class object to KERNELGLOBALS
 */

SOM_Scope void  SOMLINK xfwpsM_wpclsInitData(M_XFldWPS *somSelf)
{
    /* M_XFldWPSData *somThis = M_XFldWPSGetData(somSelf); */
    M_XFldWPSMethodDebug("M_XFldWPS","xfwpsM_wpclsInitData");

    M_XFldWPS_parent_M_WPSystem_wpclsInitData(somSelf);

    {
        // store the class object in KERNELGLOBALS
        PKERNELGLOBALS   pKernelGlobals = krnLockGlobals(__FILE__, __LINE__, __FUNCTION__);
        if (pKernelGlobals)
        {
            pKernelGlobals->fXFldWPS = TRUE;
            krnUnlockGlobals();
        }
    }
}

/* ******************************************************************
 *                                                                  *
 *   here come the XFldWPS class methods                            *
 *                                                                  *
 ********************************************************************/

/*
 *@@ wpclsQuerySettingsPageSize:
 *      this WPObject class method should return the
 *      size of the largest settings page in dialog
 *      units; if a settings notebook is initially
 *      opened, i.e. no window pos has been stored
 *      yet, the WPS will use this size, to avoid
 *      truncated settings pages.
 */

SOM_Scope BOOL  SOMLINK xfwpsM_wpclsQuerySettingsPageSize(M_XFldWPS *somSelf,
                                                          PSIZEL pSizl)
{
    BOOL brc;
    /* M_XFldWPSData *somThis = M_XFldWPSGetData(somSelf); */
    M_XFldWPSMethodDebug("M_XFldWPS","xfwpsM_wpclsQuerySettingsPageSize");

    brc = M_XFldWPS_parent_M_WPSystem_wpclsQuerySettingsPageSize(somSelf,
                                                                   pSizl);
    if (brc) {
        pSizl->cy = 170;        // this is the height of the "WPS Classes" page,
                                // which seems to be the largest in the "Workplace
                                // Shell" object
        if (doshIsWarp4())
            // on Warp 4, reduce again, because we're moving
            // the notebook buttons to the bottom
            pSizl->cy -= WARP4_NOTEBOOK_OFFSET;
    }
    return (brc);
}

/*
 *@@ wpclsQueryTitle:
 *      this WPObject class method tells the WPS the clear
 *      name of a class, which is shown in the third column
 *      of a Details view and also used as the default title
 *      for new objects of a class.
 */

SOM_Scope PSZ  SOMLINK xfwpsM_wpclsQueryTitle(M_XFldWPS *somSelf)
{
    // M_XFldWPSData *somThis = M_XFldWPSGetData(somSelf);
    M_XFldWPSMethodDebug("M_XFldWPS","xfwpsM_wpclsQueryTitle");

    return ("Workplace Shell");
}

/*
 *@@ wpclsQueryIconData:
 *      this WPObject class method builds the default
 *      icon for objects of a class (i.e. the icon which
 *      is shown if no instance icon is assigned). This
 *      apparently gets called from some of the other
 *      icon instance methods if no instance icon was
 *      found for an object. The exact mechanism of how
 *      this works is not documented.
 *
 *      We give the "Workplace Shell" object a new icon.
 */

SOM_Scope ULONG  SOMLINK xfwpsM_wpclsQueryIconData(M_XFldWPS *somSelf,
                                                   PICONINFO pIconInfo)
{
    // M_XFldWPSData *somThis = M_XFldWPSGetData(somSelf);
    M_XFldWPSMethodDebug("M_XFldWPS","xfwpsM_wpclsQueryIconData");

    if (pIconInfo)
    {
        pIconInfo->fFormat = ICON_RESOURCE;
        pIconInfo->resid   = ID_ICONWPS;
        pIconInfo->hmod    = cmnQueryMainResModuleHandle();
    }

    return (sizeof(ICONINFO));
}


